# xonox - an alternative service for legacy NOXON(tm) devices
# (c) 2022 - TillW
# Licensed to you under Affero GPL 3.0 (https://www.gnu.org/licenses/agpl-3.0.html)

class Station:
    def __init__(self, name, description, stream, id = None):
        self.id = id
        self.name = name
        self.description = description
        self.stream = stream



from .station import Station
from .station_repository import StationRepository
from .preset import Preset
from .preset_repository import PresetRepository
from .preset_service import PresetService
from .config import Config


# xonox - an alternative service for legacy NOXON(tm) devices
# (c) 2022 - TillW
# Licensed to you under Affero GPL 3.0 (https://www.gnu.org/licenses/agpl-3.0.html)

import json
from pathlib import Path
from os import path
from . import Station

class StationRepository:
    def __init__(self, config):
        self.__data = []
        self.__next_station_id = 0
        self.__config = config
        self.__read_data_from_config()

    def add(self, station):
        self.__add(station)
        self.__write_data_to_config()
        self.__save_config()

    def get(self, id):
        for x in self.__data:
            if x.id == id:
                return x
        raise KeyError()

    def get_all(self):
        return self.__data.copy()

    def remove(self, id):
        for x in self.__data:
            if x.id == id:
                self.__data.remove(x)
                self.__write_data_to_config()
                self.__save_config()
                return
        raise KeyError()

    def __add(self, station):        
        self.__update_next_station_id(station)
        self.__ensure_station_id_set(station)
        self.__data.append(station)

    def __ensure_station_id_set(self, station):
        if station.id is None:
            station.id = self.__next_station_id
            self.__next_station_id = self.__next_station_id + 1

    def __update_next_station_id(self, station):
        if station.id is not None and station.id >= self.__next_station_id:
            self.__next_station_id = station.id + 1

    def __read_data_from_config(self):
        self.__read_stations_from_config()
        self.__read_next_station_id_from_config()

    def __read_stations_from_config(self):
        if 'stations' in self.__config:
            stations = self.__config['stations']
            for station in stations:
                self.__add(Station(station['name'], station['description'], station['stream']))

    def __read_next_station_id_from_config(self):
        if 'nextStationId' in self.__config:
            try:
                self.__next_station_id = int(self.__config['nextStationId'])
            except ValueError:
                self.__next_station_id = len(self.__data)

    def __write_data_to_config(self):
        self.__config['stations'] = self.__data
        self.__config['nextStationId'] = self.__next_station_id

    def __save_config(self):
        self.__config.save()



# xonox - an alternative service for legacy NOXON(tm) devices
# (c) 2022 - TillW
# Licensed to you under Affero GPL 3.0 (https://www.gnu.org/licenses/agpl-3.0.html)

from . import Preset

class PresetRepository:
    def __init__(self, config):
        self.__data = dict()
        self.__config = config
        self.__read_presets_from_config()

    def add(self, preset):
        self.__add(preset)
        self.__write_data_to_config()
        self.__save_config()

    def get(self, group_id, preset_index):
        return self.__data[group_id][preset_index]

    def get_all(self):
        all = []
        for v in self.__data.values():
            for preset in v.values():
                all.append(preset)
        return all

    def __add(self, preset):        
        if preset.group_id not in self.__data.keys():
            self.__data[preset.group_id] = dict()        
        self.__data[preset.group_id][preset.index] = preset

    def __read_presets_from_config(self):
        if 'presets' in self.__config:
            presets = self.__config['presets']
            for preset in presets:
                self.__add(Preset(preset['group_id'], preset['index'], preset['station_id']))

    def __write_data_to_config(self):
        self.__config['presets'] = self.get_all()

    def __save_config(self):
        self.__config.save()



# xonox - an alternative service for legacy NOXON(tm) devices
# (c) 2022 - TillW
# Licensed to you under Affero GPL 3.0 (https://www.gnu.org/licenses/agpl-3.0.html)

from . import Preset, PresetRepository

class PresetService:
    def __init__(self, config):
        self.__config = config
        self.__repository = PresetRepository(config)
    
    def add_preset(self, device_id, preset_index, station_id):
        '''Adds or overwrites a preset. If the global preset list is active, 'device_id' is ignored.'''
        group_id = self.__determine_group_id(device_id)
        preset = Preset(group_id, preset_index, station_id)
        self.__repository.add(preset)

    def get_preset(self, device_id, preset_index):
        '''Gets a preset. If the global preset list is active, 'device_id' is ignored.'''
        group_id = self.__determine_group_id(device_id)
        return self.__repository.get(group_id, preset_index)

    def get_all_presets(self):
        '''Gets all presets.'''
        return self.__repository.get_all()

    def __determine_group_id(self, device_id):
        if 'settings' not in self.__config or 'useGlobalPresetList' not in self.__config['settings'] or self.__config['settings']['useGlobalPresetList']:
            return 'global'
        else:
            return device_id


# xonox - an alternative service for legacy NOXON(tm) devices
# (c) 2022 - TillW
# Licensed to you under Affero GPL 3.0 (https://www.gnu.org/licenses/agpl-3.0.html)

from functools import wraps
from collections import namedtuple
from flask import Flask, request, abort, jsonify, json, Response
from . import Config, Station, StationRepository, Preset, PresetService

# WebAPI Helpers #################
##################################
def convert_input_to(class_):
    def decorator(fn):
        '''A decorator to create an object from the JSON transferred in a request-body.'''
        @wraps(fn)
        def wrapper(*args, **kwargs):
            obj = class_(**request.get_json())
            return fn(obj)
        return wrapper
    return decorator

class ObjectToJsonStringEncoder(json.JSONEncoder):
    '''A JSON encoder that tries to serialize using 'to_json' or returns the dictionary of the object to serialize.'''
    def default(self, o):
        if hasattr(o, 'to_json'):
            return o.to_json()
        if isinstance(o, Station): # Change name of property 'stream' to 'streamUrl'
            o2 = dict(o.__dict__)
            o2['streamUrl'] = o2.pop('stream')
            return o2
        else:
            return o.__dict__

# WebAPI Initialization ##########
##################################
app = Flask(__name__)
stationRepository = None
presetService = None
stationTracker = dict() # used to track the last requested station per device to support presets/favorites
app.json_encoder = ObjectToJsonStringEncoder

# Management API #################
##################################
CreateStationDto = namedtuple('CreateStationDto', 'name description streamUrl')

@app.route('/station', methods=['post'])
@convert_input_to(CreateStationDto)
def create_station(dto):
    station = Station(dto.name, dto.description, dto.streamUrl)
    stationRepository.add(station)
    return Response(status=201, headers={'Location': '/station/' + str(station.id)}, content_type='application/json')

@app.route('/station', methods=['get'])
def get_station_list():
    return jsonify(stationRepository.get_all())

@app.route('/station/<int:id>', methods=['get'])
def get_station(id):
    try:
        return jsonify(stationRepository.get(id))
    except KeyError:
        return abort(404)

@app.route('/station/<int:id>', methods=['delete'])
def delete_station(id):
    try:
        stationRepository.remove(id)
        return Response(status=204)
    except KeyError:
        return abort(404)

@app.route('/settings', methods=['post'])
@convert_input_to(namedtuple('WriteSettingsDto', 'useGlobalPresetList'))
def read_config(dto):
    global config
    settings = {'useGlobalPresetList': bool(dto.useGlobalPresetList) }
    config['settings'] = settings 
    config.save()
    return Response(status=204)

@app.route('/settings', methods=['get'])
def read_settings():
    global config
    if 'settings' in config:
        return jsonify(config['settings'])
    else:
        return jsonify({})

# NOXON(tm) API ##################
##################################
@app.route('/setupapp/fs/asp/BrowseXML/loginXML.asp')
def get_root_menu():
    if (__get_device_id(request) is None):
        return '<EncryptedToken>a6703ded78821be5</EncryptedToken>'
    else:
        sorted_stations = sorted(stationRepository.get_all(), key=lambda x: x.name)
        return __create_station_list(sorted_stations, request.host_url)

@app.route('/setupapp/fs/asp/BrowseXML/Search.asp')
def search_station():
    requestedStationId = int(request.args.get('Search'))
    try:
        station = stationRepository.get(requestedStationId)
        __track_station(station, request)
        return __create_station_list([station], request.host_url)
    except IndexError:
        return abort(404)

@app.route('/Favorites/AddPreset.aspx')
def add_preset():
    device_id, preset_index = __get_device_and_preset_index(request)
    if device_id in stationTracker.keys():
        station_id = stationTracker[device_id]
        presetService.add_preset(device_id, preset_index, station_id)
        result = '<?xml version="1.0" encoding="iso-8859-1" standalone="yes"?><ListOfItems><ItemCount>-1</ItemCount><Item><ItemType>Message</ItemType><Message>Preset set</Message></Item></ListOfItems>'
    else:
        result = '<?xml version="1.0" encoding="iso-8859-1" standalone="yes"?><ListOfItems><ItemCount>-1</ItemCount>'
        result = result + '<Item><ItemType>Message</ItemType><Message>Preset not created. Reselect the station to preset and then try again.</Message></Item>'
        result = result + '</ListOfItems>'
    return result

@app.route('/Favorites/GetPreset.aspx')
def get_preset():
    group_id, preset_index = __get_device_and_preset_index(request)
    try:
        preset = presetService.get_preset(group_id, preset_index)
        station = stationRepository.get(preset.station_id)
        return __create_station_list([station], request.host_url)
    except KeyError:
        return '<?xml version="1.0" encoding="iso-8859-1" standalone="yes"?><ListOfItems><ItemCount>-1</ItemCount><Item><ItemType>Message</ItemType><Message>Preset not found</Message></Item></ListOfItems>'

@app.route('/noOp')
def no_op():
    return ''

def __create_station_list(stations, baseUri):
    result = '<?xml version="1.0" encoding="iso-8859-1" standalone="yes"?><ListOfItems><ItemCount>-1</ItemCount>'
    for station in stations:
        result = result + __station_to_xml(station, baseUri)
    result = result + '</ListOfItems>'
    return result

def __station_to_xml(station, baseUri):
    return '<Item><ItemType>Station</ItemType><StationId>' + str(station.id) + '</StationId><StationName>' + station.name + '</StationName><StationUrl>' + station.stream + '</StationUrl><StationDesc>' + station.description + '</StationDesc><StationFormat>Public</StationFormat><StationLocation>n/a</StationLocation><StationBandWidth>128</StationBandWidth><StationMime>MP3</StationMime><Relia>1</Relia><Bookmark>' + baseUri + '/noOp</Bookmark><Logo>' + baseUri + '/noOp</Logo></Item>'

def __track_station(station, request):    
    device_id = __get_device_id(request)
    stationTracker[device_id] = station.id

def __get_device_id(request):
    return str(request.args.get('mac'))

def __get_preset_index(request):
    return int(request.args.get('id'))

def __get_device_and_preset_index(request):
    return (__get_device_id(request), __get_preset_index(request))

# Application ####################
##################################
def run(host, config_directory):
    global stationRepository
    global presetService
    global config
    config = Config(config_directory)
    stationRepository = StationRepository(config)
    presetService = PresetService(config)
    app.run(host=host, port=80)


# xonox - an alternative service for legacy NOXON(tm) devices
# (c) 2022 - TillW
# Licensed to you under Affero GPL 3.0 (https://www.gnu.org/licenses/agpl-3.0.html)

class Preset:
    def __init__(self, group_id, preset_index, station_id):
        self.group_id = group_id
        self.index = preset_index
        self.station_id = station_id



# xonox - an alternative service for legacy NOXON(tm) devices
# (c) 2022 - TillW
# Licensed to you under Affero GPL 3.0 (https://www.gnu.org/licenses/agpl-3.0.html)

import json
from pathlib import Path
from os import path

class Config:
    def __init__(self, config_directory):
        self.__data = dict()
        if config_directory is None:
            config_directory = Path.home()
        self.__config_path = path.join(config_directory, 'xonox.conf')
        self.__load_data_from_file()

    def __getitem__(self, key):
        return self.__data[key]

    def __setitem__(self, key, value):
        self.__data[key] = value

    def __contains__(self, key):
        return self.__data.__contains__(key)

    def save(self):
        with open(self.__config_path, 'w+') as file:
            json.dump(self.__data, file, default=lambda o: o.__dict__, sort_keys=False, indent=2)

    def __load_data_from_file(self):
        try:
            with open(self.__config_path, 'r') as file:
                self.__data = json.load(file)
        except FileNotFoundError:
            pass


# xonox - an alternative service for legacy NOXON(tm) devices
# (c) 2022 - TillW
# Licensed to you under Affero GPL 3.0 (https://www.gnu.org/licenses/agpl-3.0.html)

from argparse import ArgumentParser
from xonox import server

if __name__ == "__main__":
    parser = ArgumentParser(description = 'An alternative infrastructure-service for legacy Noxon(tm) devices.')
    parser.add_argument('--host', default='0.0.0.0', required=False, help='IP address to bind to. If omitted, xonox bind to all available interfaces.')
    parser.add_argument('--config-dir', default=None, dest='config_dir', help='Directory that contains the configuration. If omitted, the user\'s home directory is used.')
    
    args = parser.parse_args()
    server.run(args.host, args.config_dir)